"""
Module for Tracking and Displaying Usage Statistics of LLM Completions and Transcriptions.

This module provides classes and methods to track usage statistics for language model completions
and transcriptions, compute associated costs, and display the information using Streamlit expanders.
It integrates with DynamoDB for persistent storage of daily usage data.

The cost calucation done here is an estimate at best, and misleading at worst. It doesn't currently
have any way to represent a change in model cost, as the only thing tracked is the actual usage itself.
If it is necessary to be able to compute the changing cost of a model over time, that information will
need to be tracked separately.

Classes:
    UsageStats:
        - Tracks input/output tokens, completions, and transcription data by model.
        - Computes costs per model based on usage and model pricing.
        - Provides methods to render usage statistics as Streamlit expanders.

    GlobalUsageTracker (UsageStats, SingletonResource):
        - A singleton resource that globally tracks usage statistics across sessions.

    DailyUsageTracking (UsageStats, DynamoDbResource):
        - Manages daily usage statistics with DynamoDB integration.
        - Provides methods to retrieve or create daily tracking records.
        - Supports querying usage data over a date range.

Usage:
    - Instantiate `UsageStats` to track and compute costs for LLM usage.
    - Use `GlobalUsageTracker` for application-wide usage tracking.
    - Use `DailyUsageTracking` to store and retrieve daily usage statistics from DynamoDB.

Examples:
    # Compute and display completion costs
    usage_stats = UsageStats(...)
    usage_stats.render_completion_cost_as_expander()

    # Get today's usage tracking record from DynamoDB
    memory = DynamoDbMemory(...)
    daily_usage = DailyUsageTracking.get_for_today(memory)

"""

import datetime
from typing import Optional, Union

import pandas as pd
import streamlit as st
from boto3.dynamodb.conditions import Key
from pydantic import BaseModel, Field
from simplesingletable import DynamoDbMemory, DynamoDbResource
from simplesingletable.extras.singleton import SingletonResource

from supersullytools.llm.completions import (
    CompletionModel,
    CompletionResponse,
    ImagePromptMessage,
    PromptAndResponse,
    PromptMessage,
)


class UsageStats(BaseModel):
    # SECTION 1: llm completion tracking / helpers
    input_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    output_tokens_by_model: dict[str, int] = Field(default_factory=dict)
    completions_by_model: dict[str, int] = Field(default_factory=dict)

    def compute_cost_per_model(self, model_classes: Optional[list[CompletionModel]] = None):
        if model_classes is None:
            from supersullytools.llm.completions import ALL_MODELS

            model_classes = ALL_MODELS

        # Build a mapping from model identifiers to model classes
        model_class_by_id = {mc.llm_id: mc for mc in model_classes}
        model_class_by_name = {mc.llm: mc for mc in model_classes}
        model_class_lookup = {**model_class_by_id, **model_class_by_name}

        results = {}
        for model in self.completions_by_model:
            model_class = model_class_lookup.get(model)
            if not model_class:
                continue  # Skip if model_class is not found
            input_tokens = self.input_tokens_by_model.get(model, 0)
            output_tokens = self.output_tokens_by_model.get(model, 0)
            input_cost = input_tokens / 1000 * model_class.input_price_per_1k
            output_cost = output_tokens / 1000 * model_class.output_price_per_1k
            results[model] = input_cost + output_cost
        return results

    def render_completion_cost_as_expander(
        self, header="Total Completions: ", model_classes: Optional[list[CompletionModel]] = None, total_in_header=True
    ):
        stats_df = pd.DataFrame(
            {
                "count": self.completions_by_model,
                "input_tokens": self.input_tokens_by_model,
                "output_tokens": self.output_tokens_by_model,
                "cost": self.compute_cost_per_model(model_classes),
            }
        )
        if not stats_df["count"].count():
            completions_label = "No Completions"
            total_cost = 0.0
        else:
            total_count = int(stats_df["count"].sum())
            total_cost = round(stats_df["cost"].sum(), 4)
            completions_label = f"{header}{total_count}"
            if total_in_header:
                completions_label += f"\n\nApprox Cost: ${total_cost}"

        with st.expander(completions_label):
            if not total_in_header and total_cost > 0:
                st.info(f"Approx Completions Cost: ${total_cost}")
            st.dataframe(stats_df.fillna(0))

    # SECTION 2: transcription tracking / helpers
    transcripts_by_model: dict[str, int] = Field(default_factory=dict)
    seconds_transcribed_by_model: dict[str, int] = Field(default_factory=dict)

    def get_minutes_transcribed_by_model(self) -> dict[str, float]:
        return {k: v / 60 for k, v in self.seconds_transcribed_by_model.items()}

    def compute_transcript_cost_per_model(self, model_cost_by_minute: dict[str, float]):
        minutes_transcribed = self.get_minutes_transcribed_by_model()
        results = {}
        for model in self.transcripts_by_model:
            minutes = minutes_transcribed.get(model, 0.0)
            price_per_minute = model_cost_by_minute.get(model, 0.0)
            results[model] = minutes * price_per_minute
        return results

    def render_transcript_cost_as_expander(self, model_cost_by_minute: dict[str, float]):
        cost_per_model = self.compute_transcript_cost_per_model(model_cost_by_minute)
        stats_df = pd.DataFrame(
            {
                "count": self.transcripts_by_model,
                "minutes": self.get_minutes_transcribed_by_model(),
                "cost": cost_per_model,
            }
        )
        if not stats_df["count"].count():
            completions_label = "No Transcripts"
        else:
            total_count = int(stats_df["count"].sum())
            total_minutes = round(stats_df["minutes"].sum(), 2)
            total_cost = round(stats_df["cost"].sum(), 4)
            completions_label = f"Total Transcripts: {total_count} ({total_minutes} min)\n\nApprox Cost: ${total_cost}"

        with st.expander(completions_label):
            st.dataframe(stats_df.fillna(0))


class GlobalUsageTracker(SingletonResource, UsageStats):
    pass


class DailyUsageTracking(DynamoDbResource, UsageStats):
    day: str

    @classmethod
    def get_for_today(cls, memory: DynamoDbMemory, today: Optional[datetime.date] = None, consistent_read: bool = True):
        today = today or datetime.date.today()
        key = today.strftime("%Y%m%d")
        if not (existing := memory.get_existing(key, data_class=cls, consistent_read=consistent_read)):
            return memory.create_new(cls, {"day": today.isoformat()}, override_id=key)
        return existing

    @classmethod
    def get_by_date_range(
        cls, memory: DynamoDbMemory, first_date: datetime.date, last_date: datetime.date
    ) -> list["DailyUsageTracking"]:
        first_day = first_date.isoformat()
        last_day = (last_date + datetime.timedelta(days=1)).isoformat()
        return memory.paginated_dynamodb_query(
            key_condition=(Key("gsitype").eq(cls.__name__) & Key("gsitypesk").between(first_day, last_day)),
            index_name="gsitype",
            resource_class=cls,
        )

    def db_get_gsitypesk(self) -> str:
        # override the base method; use created_at instead of updated_at for the type sort key,
        # to enable easy queries for a range of dates off this index (see the classmethod `get_by_date_range`)
        return self.created_at.isoformat()


class TopicUsageTracking(DynamoDbResource, UsageStats):
    topic: str

    @classmethod
    def get_for_topic(cls, memory: DynamoDbMemory, topic: str, consistent_read: bool = True) -> "TopicUsageTracking":
        if not (existing := memory.get_existing(topic, data_class=cls, consistent_read=consistent_read)):
            return memory.create_new(cls, {"topic": topic}, override_id=topic)
        return existing


class SessionUsageTracking(UsageStats):
    completions: list[PromptAndResponse] = Field(default_factory=list)

    def reset(self):
        self.completions = []
        self.input_tokens_by_model = {}
        self.output_tokens_by_model = {}
        self.completions_by_model = {}


TrackerTypes = Union[UsageStats, GlobalUsageTracker, DailyUsageTracking, TopicUsageTracking]


class CompletionTracker(object):
    def __init__(self, memory: DynamoDbMemory, trackers: list[TrackerTypes]):
        self.memory = memory
        self.trackers = trackers

    def track_completion(
        self, model: CompletionModel, prompt: list[PromptMessage | ImagePromptMessage], completion: CompletionResponse
    ):
        llm_to_track = model.llm
        for tracker in self.trackers:
            if isinstance(tracker, SessionUsageTracking):
                if llm_to_track in tracker.input_tokens_by_model:
                    tracker.completions_by_model[llm_to_track] += 1
                    tracker.input_tokens_by_model[llm_to_track] += completion.input_tokens
                    tracker.output_tokens_by_model[llm_to_track] += completion.output_tokens
                else:
                    tracker.completions_by_model[llm_to_track] = 1
                    tracker.input_tokens_by_model[llm_to_track] = completion.input_tokens
                    tracker.output_tokens_by_model[llm_to_track] = completion.output_tokens
                tracker.completions.append(PromptAndResponse(prompt=prompt, response=completion))

            if isinstance(tracker, DynamoDbResource):
                self.memory.increment_counter(tracker, f"completions_by_model.{llm_to_track}")
                self.memory.increment_counter(tracker, f"input_tokens_by_model.{llm_to_track}", completion.input_tokens)
                self.memory.increment_counter(
                    tracker, f"output_tokens_by_model.{llm_to_track}", completion.output_tokens
                )



import os
import time

import streamlit as st
from logzero import logger
from simplesingletable import DynamoDbMemory

from supersullytools.llm.agent import ChatAgent
from supersullytools.llm.agent_tools.duckduckgo import get_ddg_tools
from supersullytools.llm.completions import CompletionHandler
from supersullytools.llm.trackers import CompletionTracker, DailyUsageTracking, GlobalUsageTracker, SessionUsageTracking
from supersullytools.streamlit.chat_agent_utils import ChatAgentUtils


@st.cache_resource
def get_memory() -> DynamoDbMemory:
    return DynamoDbMemory(logger=logger, table_name=os.environ.get("DYNAMODB_TABLE"))


@st.cache_resource
def get_completion_handler() -> CompletionHandler:
    memory = get_memory()
    trackers = get_trackers()
    completion_tracker = CompletionTracker(memory=memory, trackers=list(trackers))
    return CompletionHandler(
        logger=logger, debug_output_prompt_and_response=False, completion_tracker=completion_tracker
    )


@st.cache_resource
def get_session_usage_tracker() -> SessionUsageTracking:
    return SessionUsageTracking()


def get_trackers() -> tuple[GlobalUsageTracker, DailyUsageTracking, SessionUsageTracking]:
    memory = get_memory()
    global_tracker = GlobalUsageTracker.ensure_exists(memory)
    todays_tracker = DailyUsageTracking.get_for_today(memory)
    return global_tracker, todays_tracker, get_session_usage_tracker()


@st.cache_resource
def get_agent() -> ChatAgent:
    tool_profiles = {"all": [] + get_ddg_tools()}
    return ChatAgent(
        agent_description="You are a helpful assistant.",
        logger=logger,
        completion_handler=get_completion_handler(),
        tool_profiles=tool_profiles,
    )


def main():
    with st.sidebar:
        model = ChatAgentUtils.select_llm(get_completion_handler(), label="LLM to use")
    st.title("AI Chat Agent Testing")

    def _agent():
        agent = get_agent()
        agent.default_completion_model = model
        return agent

    agent = _agent()
    agent_utils = ChatAgentUtils(agent)

    if "image_key" not in st.session_state:
        st.session_state.image_key = 1
        st.session_state.upload_images = []

    image = st.sidebar.file_uploader("Image", type=["png", "jpg"], key=f"image-upload-{st.session_state.image_key}")
    if image and st.sidebar.button("Add image to msg"):
        st.session_state.image_key += 1
        st.session_state.upload_images.append(image)
        time.sleep(0.01)
        st.rerun()

    if st.session_state.upload_images:
        with st.sidebar.expander("Pending Images", expanded=True):
            for image in st.session_state.upload_images:
                st.image(image)

    chat_msg = st.chat_input()

    with st.sidebar.expander("Chat Config", expanded=True):
        include_function_calls = st.sidebar.toggle("Show function calls", True)

    agent_utils.display_chat_and_run_agent(include_function_calls)

    if chat_msg:
        if agent_utils.add_user_message(chat_msg, st.session_state.upload_images):
            if st.session_state.upload_images:
                # clearing out the upload_images immediately causes weird IO errors, so
                # just push to another key and overwrite it later
                st.session_state.uploaded_images = st.session_state.upload_images
                st.session_state.upload_images = []
            time.sleep(0.01)
            st.rerun()

    global_tracker, daily_tracker, session_tracker = get_trackers()
    with st.sidebar.container(border=True):
        st.subheader("Total Usage")
        global_tracker.render_completion_cost_as_expander()
        st.subheader("Daily Usage")
        daily_tracker.render_completion_cost_as_expander()
        st.subheader("Session Usage")
        session_tracker.render_completion_cost_as_expander()


if __name__ == "__main__":
    main()



from typing import Sequence

import streamlit as st


def check_or_x(value: bool) -> str:
    return "✅" if value else "❌"


def resettable_tabs(name: str, tabs=Sequence[str], session_key_prefix: str = "resettable_tabs_") -> str:
    key = f"{session_key_prefix}{name}"
    for x in range(st.session_state.get(key, 0)):
        st.empty()
    return st.tabs(tabs)


def reset_tab_group(name: str, session_key_prefix: str = "resettable_tabs_"):
    key = f"{session_key_prefix}{name}"
    current = st.session_state.get(key, 0)
    st.session_state[key] = current + 1


def flash_message_after_reload(
    msg: str, toast=False, flash_msgs_session_key="flash_msgs", flash_toasts_session_session_key="flash_toast"
):
    key = flash_toasts_session_session_key if toast else flash_msgs_session_key
    if key not in st.session_state:
        st.session_state[key] = [msg]
    else:
        st.session_state[key].append(msg)


def display_flash_msgs(
    flash_msgs_session_key="flash_msgs", flash_toasts_session_session_key="flash_toast", container=None
):
    if flash_msgs := st.session_state.get(flash_msgs_session_key):
        for msg in flash_msgs:
            if container:
                with container:
                    st.write(msg)
            else:
                st.write(msg)
        st.session_state[flash_msgs_session_key] = []
    if toast_msgs := st.session_state.get(flash_toasts_session_session_key):
        for msg in toast_msgs:
            st.toast(msg)
        st.session_state[flash_toasts_session_session_key] = []



import streamlit as st

from supersullytools.streamlit.misc import check_or_x
from supersullytools.streamlit.paginator import item_paginator

sample_data = [
    {"name": "Item1", "category": "Electronics", "price": 299.99, "rating": 4.5, "in_stock": True},
    {"name": "Item2", "category": "Books", "price": 15.99, "rating": 4.8, "in_stock": False},
    {"name": "Item3", "category": "Clothing", "price": 49.99, "rating": 3.9, "in_stock": True},
    {"name": "Item4", "category": "Groceries", "price": 3.49, "rating": 4.2, "in_stock": True},
]


def display_sample_item(item_idx):
    item = sample_data[item_idx]
    st.subheader(item["name"])
    st.caption(f"category={item['category']}")
    metric_cols = iter(st.columns(3))
    with next(metric_cols):
        st.metric("Price", item["price"])
    with next(metric_cols):
        st.metric("Rating", item["rating"])
    with next(metric_cols):
        st.metric("In-Stock", check_or_x(item["in_stock"]))


def main():
    st.write("Use left/right arrows to navigate between items.")
    item_names = [x["name"] for x in sample_data]
    item_paginator(
        title="Available Items",
        items=item_names,
        display_item_names=True,
        item_handler_fn=display_sample_item,
        enable_keypress_nav=True,
    )


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Iterable

from supersullytools.utils import fuzzy_search


class TestScoreStringSimilarity:
    def test_find_most_similar(self):
        # source string is first, term is second. The score is related to adjustments for the term to match the source
        assert fuzzy_search.score_string_similarity("Cat", "Cat", weights=(1, 1, 1)) == 0

        # one addition
        assert fuzzy_search.score_string_similarity("Cat", "at", weights=(1, 2, 3)) == 1
        # one deletion
        assert fuzzy_search.score_string_similarity("Cat", "Cate", weights=(1, 2, 3)) == 2
        # one change
        assert fuzzy_search.score_string_similarity("Cat", "C1t", weights=(1, 2, 3)) == 3

        # two additions
        assert fuzzy_search.score_string_similarity("Cat", "a", weights=(1, 9, 9)) == 2
        # two deletions
        assert fuzzy_search.score_string_similarity("Cat", "Crate", weights=(9, 2, 9)) == 4
        # two changes
        assert fuzzy_search.score_string_similarity("Cat", "C1a", weights=(9, 9, 3)) == 6

        # two changes, but a delete/add are cheaper
        assert fuzzy_search.score_string_similarity("Cat", "C1a", weights=(1, 1, 3)) == 2

        # 3 changes
        assert fuzzy_search.score_string_similarity("Cat", "ZOO", weights=(100, 100, 3)) == 9
        # 3 changes, but cheaper to delete and add everything
        assert fuzzy_search.score_string_similarity("Cat", "ZOO", weights=(1, 1, 3)) == 6

    def test_supports_similarity_protocol(self):
        @dataclass
        class MyTest:
            value: str = "Ziya Brinhilde"

            def get_full_value(self) -> str:
                return self.value

            def get_search_terms(self) -> Iterable[str]:
                val = self.value.lower()
                first, last = val.split(" ")
                return [first, last, first + last]

        test_obj = MyTest()
        assert fuzzy_search.score_string_similarity(test_obj, "ziya", (1, 10, 10)) == 0
        assert fuzzy_search.score_string_similarity(test_obj, "brinhilde", (1, 10, 10)) == 0

        # our search terms specifically .lower() the values, so searching with uppercase is a "change" score
        assert fuzzy_search.score_string_similarity(test_obj, "Ziya", (1, 10, 10)) == 10
        assert fuzzy_search.score_string_similarity(test_obj, "zeeba", (1, 10, 10)) == 30
        assert fuzzy_search.score_string_similarity(test_obj, "hilde", (1, 10, 10)) == 4
        assert fuzzy_search.score_string_similarity(test_obj, "zyaBrhile", (1, 10, 10)) == 14


class TestGetTopNResults:
    def test_basic(self):
        search_these = ["Cat", "Dog", "Banana", "Bandana"]

        results = fuzzy_search.get_top_n_results(search_these, "anana")
        print(results)
        assert results[:2] == ["Banana", "Bandana"]

    def test_supports_similarity_protocol(self):
        @dataclass
        class MyTest:
            value: str = "Ziya Brinhilde"

            def get_full_value(self) -> str:
                return self.value

            def get_search_terms(self) -> Iterable[str]:
                val = self.value.lower()
                first, last = val.split(" ")
                return [first, last, first + last]

        results = fuzzy_search.get_top_n_results([MyTest()], "ziya")
        assert results == [fuzzy_search.MatchedString(score=0, value="Ziya Brinhilde", match_term="ziya")]

        results = fuzzy_search.get_top_n_results([MyTest()], "Ziya")
        assert results == [fuzzy_search.MatchedString(score=10, value="Ziya Brinhilde", match_term="ziya")]

        results = fuzzy_search.get_top_n_results([MyTest(), "Zeeba"], "zeyaa")
        assert results == [
            fuzzy_search.MatchedString(score=20, value="Ziya Brinhilde", match_term="ziya"),
            fuzzy_search.MatchedString(score=30, value="Zeeba", match_term="Zeeba"),
        ]
        results = fuzzy_search.get_top_n_results([MyTest(), "Zeeba"], "hilde")
        assert results == [fuzzy_search.MatchedString(score=4, value="Ziya Brinhilde", match_term="brinhilde")]



"""
streamlit_media_manager.py

This Streamlit app provides a demo interface for managing media files stored in an Amazon S3 bucket,
with metadata stored in Amazon DynamoDB. It includes features for uploading, retrieving, generating
previews, and deleting media files, with optional gzip compression and encryption support for efficient
and secure storage of compressible data like CSV and JSON files.

Functions:
    setup_media_manager() -> MediaManager:
        Sets up and returns a MediaManager instance for managing media files.

    detect_media_type(file_name: str) -> Optional[MediaType]:
        Detects the media type of a file based on its MIME type.

    display_content(contents, media_type: MediaType):
        Displays the content in the Streamlit app based on the media type.

    upload_media_dialog():
        Handles the file upload process, including optional gzip compression and encryption, and generates previews.

    get_preview_image_base64(media_id: str) -> str:
        Retrieves the preview image for a given media ID and encodes it in base64 format.

Streamlit Sections:
    - Upload Media: Allows users to upload media files with optional gzip compression and encryption, and preview generation.
    - Recent Uploads: Displays a table of recent uploads with preview images and options to delete selected files.
    - Retrieve Metadata: Retrieves and displays the metadata for a specified media ID.
    - Retrieve Content: Retrieves and displays the content of a specified media ID, with a download option.
    - Retrieve Preview: Retrieves and displays the preview of a specified media ID.
    - Delete Media: Deletes the specified media file and its metadata from S3 and DynamoDB.

Dependencies:
    - base64: For encoding preview images in base64 format.
    - mimetypes: For detecting MIME types of files.
    - os: For accessing environment variables.
    - pandas: For creating and manipulating DataFrames.
    - streamlit: For building the web app interface.
    - logzero: For logging information and errors.
    - simplesingletable: For interacting with DynamoDB.
    - supersullytools.utils.media_manager: For managing media files and their metadata.
    - cryptography: For generating and using encryption keys (optional).

Usage:
    1. Set up the required environment variables:
        - DYNAMODB_TABLE: The name of the DynamoDB table.
        - S3_BUCKET: The name of the S3 bucket.
        - (optional) S3_MEDIA_PREFIX: A prefix within the bucket to use.

    2. Run the Streamlit app:
        streamlit run streamlit_media_manager.py

    3. Use the app interface to upload, view, retrieve, and delete media files.
"""

import base64
import mimetypes
import os
import secrets
from io import BytesIO
from typing import Optional

import pandas as pd
import streamlit as st
from logzero import logger
from simplesingletable import DynamoDbMemory

from supersullytools.utils.media_manager import MediaManager, MediaType, generate_text_image

st.set_page_config(layout="wide")


def setup_media_manager():
    # Placeholder function to set up the MediaManager
    # Replace with your actual implementation
    dynamodb_memory = DynamoDbMemory(logger=logger, table_name=os.environ.get("DYNAMODB_TABLE"))
    return MediaManager(
        bucket_name=os.environ.get("S3_BUCKET"),
        logger=logger,
        dynamodb_memory=dynamodb_memory,
        global_prefix=os.environ.get("S3_MEDIA_PREFIX") or "media-manager-testing",
    )


def detect_media_type(file_name: str) -> Optional[MediaType]:
    mime_type, _ = mimetypes.guess_type(file_name)
    if mime_type:
        if mime_type.startswith("image"):
            return MediaType.image
        elif mime_type.startswith("audio"):
            return MediaType.audio
        elif mime_type.startswith("video"):
            return MediaType.video  # Assuming "video" is a valid media type in your MediaManager
        elif mime_type == "application/pdf":
            return MediaType.pdf
        elif mime_type.startswith("text"):
            return MediaType.text
    return None


def display_content(contents, media_type: MediaType):
    if media_type == "image":
        st.image(contents, caption="Image")
    elif media_type == "audio":
        st.audio(contents)
    elif media_type == "video":
        st.video(contents)
    else:
        st.write("Unable to display contents")


# Initialize the media manager
media_manager = setup_media_manager()

st.title("Media Manager Streamlit App")

# Key generation section
st.header("Key Generation")
if st.button("Generate New Encryption Key"):
    # Generate a random secret key (AES256 needs 32 bytes)
    new_key = base64.urlsafe_b64encode(secrets.token_bytes(32))
    st.code(new_key.decode(), language="text")

# Global encryption key
encryption_key = st.text_input("Global Encryption Key (optional)", type="password", key="global_encryption_key")
if encryption_key:
    encryption_key = base64.urlsafe_b64decode(encryption_key.encode())


@st.experimental_dialog("Upload Media", width="large")
def upload_media_dialog():
    # File upload section
    st.header("Upload Media")
    uploaded_file = st.file_uploader("Choose a file")

    default_media_type = None
    if uploaded_file:
        media_types = list(MediaType)
        st.write(mimetypes.guess_type(uploaded_file.name)[0])
        try:
            default_media_type = media_types.index(detect_media_type(uploaded_file.name))
        except ValueError:
            pass

    media_type: MediaType = st.selectbox(
        "Select media type", options=list(MediaType), index=default_media_type, format_func=lambda x: x.value
    )

    if uploaded_file is not None and media_type:
        st.write("Filename:", uploaded_file.name)
        bytes_data = uploaded_file.read()
        file_obj = BytesIO(bytes_data)
        gzip_content = st.checkbox("Gzip content before storage")
        encrypt_content = st.checkbox("Encrypt content")
        encrypt_preview = st.checkbox("Encrypt preview")
        try:
            preview = media_manager.generate_preview(uploaded_file, media_type)
            st.image(preview, media_type)
        except Exception as e:
            st.error(f"Failed to generate preview for file: {str(e)}")

        if st.button("Upload"):
            try:
                metadata = media_manager.upload_new_media(
                    uploaded_file.name,
                    media_type,
                    file_obj,
                    use_gzip=gzip_content,
                    encryption_key=encryption_key if encrypt_content else "",
                    encrypt_preview=encrypt_preview,
                )
                st.experimental_rerun()

            except Exception as e:
                st.error(f"Failed to upload file: {str(e)}")
            else:
                return metadata


if st.button("Upload Media"):
    upload_media_dialog()

st.header("Recent Uploads")

data = []
previews = []


@st.cache_data(persist=True)
def get_preview_image_base64(media_id: str, encryption_key: Optional[bytes] = None) -> str:
    preview_content = media_manager.retrieve_media_preview(media_id, encryption_key=encryption_key)
    return base64.b64encode(preview_content).decode("utf-8")


@st.cache_data(ttl=15)
def list_media():
    return media_manager.list_available_media(num=50, oldest_first=False)


for media in list_media():
    if media.preview_encrypted:
        if encryption_key:
            previews.append(get_preview_image_base64(media.resource_id, encryption_key))
        else:
            previews.append(base64.b64encode(generate_text_image("Encryption Key Required")).decode("utf-8"))
    else:
        previews.append(get_preview_image_base64(media.resource_id))

    media_dict = media.model_dump(
        mode="json",
        exclude={
            "created_at",
            "updated_at",
            "file_size_bytes",
            "preview_size_bytes",
            "storage_size_bytes",
            "preview_storage_size_bytes",
        },
    )
    data.append(media_dict)


# Convert list of dictionaries to DataFrame
df = pd.DataFrame(data)

# Add a column for the preview images
df["Preview"] = [f"data:image/jpeg;base64,{preview}" for preview in previews]

# Display the DataFrame with custom column configuration
selected = st.dataframe(
    df,
    column_config={
        "Preview": st.column_config.ImageColumn(
            label="Preview Image",
            help="Preview of the uploaded media",
            width="small",
        ),
    },
    use_container_width=True,
    on_select="rerun",
    selection_mode="multi-row",
)

if work_on_rows := selected["selection"]["rows"]:
    selected_media_ids = [df.iloc[x]["resource_id"] for x in work_on_rows]
    if st.button("Delete selected", type="primary"):
        for delete_media_id in selected_media_ids:
            try:
                preview_metadata = media_manager.delete_media(delete_media_id)
                st.info(f"Deleted {delete_media_id}")
            except Exception as e:
                st.error(f"Failed to delete media: {str(e)}")

# Retrieve metadata section
st.header("Retrieve Metadata")
media_id = st.text_input("Enter media ID to retrieve metadata")
if st.button("Retrieve Metadata"):
    try:
        metadata = media_manager.retrieve_metadata(media_id)
        st.json(metadata.dict())
    except Exception as e:
        st.error(f"Failed to retrieve metadata: {str(e)}")

# Retrieve content section
st.header("Retrieve Content")
media_id_content = st.text_input("Enter media ID to retrieve content")
if st.button("Retrieve Content"):
    try:
        contents_metadata, contents = media_manager.retrieve_media_metadata_and_contents(
            media_id_content, encryption_key=encryption_key
        )
        st.write(contents_metadata.src_filename)
        display_content(contents, contents_metadata.media_type)
        st.download_button("Download Content", data=contents, file_name=contents_metadata.src_filename)
    except Exception as e:
        st.error(f"Failed to retrieve content: {str(e)}")

# Retrieve preview section
st.header("Retrieve Preview")
media_id_preview = st.text_input("Enter media ID to retrieve preview")
if st.button("Retrieve Preview"):
    try:
        preview_metadata = media_manager.retrieve_metadata(media_id_preview)
        preview_contents = media_manager.retrieve_media_preview(media_id_preview, encryption_key=encryption_key)
        st.image(preview_contents, preview_metadata.media_type)
    except Exception as e:
        st.error(f"Failed to retrieve preview: {str(e)}")

# Delete media section
st.header("Delete Media")
delete_media_id = st.text_input("Enter media ID to delete; this cannot be undone!")
if st.button("Delete Media", type="primary"):
    try:
        media_manager.delete_media(delete_media_id)
        st.info("Deleted")
    except Exception as e:
        st.error(f"Failed to delete media: {str(e)}")



from dataclasses import dataclass, field
from logging import Logger
from typing import Optional, Type, TypeVar

import openai
from openai import Client
from openai.types import ModerationCreateResponse
from openai.types.chat import ChatCompletion
from openai.types.chat.completion_create_params import Function
from pydantic import BaseModel


class ChatTool(BaseModel):
    pass

    @classmethod
    def to_chat_function(cls) -> Function:
        return Function(name=cls.__name__, description=cls.__doc__, parameters=cls.model_json_schema())


_T = TypeVar("_T", bound=ChatTool)


# NOTE: this is now "legacy" and I'm using the supersullytools.llm.completions CompletionHandler class now


@dataclass
class ChatSession:
    initial_system_message: Optional[str] = None
    reinforcement_system_msg: Optional[str] = None
    default_tools: list[Type[_T]] = None
    history: list = field(default_factory=list)
    log: Logger = None
    model: str = "gpt-3.5-turbo-0125"
    openai: Client = None

    def __post_init__(self):
        if self.log is None:
            from logzero import logger

            self.log = logger

    def user_says(self, message):
        self.history.append({"role": "user", "content": message})

    def system_says(self, message):
        self.history.append({"role": "system", "content": message})

    def assistant_says(self, message):
        self.history.append({"role": "assistant", "content": message})

    def openai_client(self) -> Client:
        return self.openai or openai

    @classmethod
    def list_gpt_models(cls):
        return sorted([x.id for x in openai.models.list() if "gpt" in x.id and "instruct" not in x.id])

    def get_ai_response(
        self,
        initial_system_msg_override: str = None,
        reinforcement_system_msg_override: str = None,
        tools_override: list[type[_T]] = None,
        force_tool: Optional[str | bool] = None,
    ) -> ChatCompletion:
        initial_system_msg = initial_system_msg_override or self.initial_system_message
        reinforcement_system_msg = reinforcement_system_msg_override or self.reinforcement_system_msg
        tools = tools_override or self.default_tools

        chat_history = self.history[:]
        # add the initial system message describing the AI's role
        if initial_system_msg:
            chat_history.insert(0, {"role": "system", "content": initial_system_msg})

        if reinforcement_system_msg:
            chat_history.append({"role": "system", "content": reinforcement_system_msg})
        self.log.info("Generating AI ChatCompletion")
        self.log.debug(chat_history)
        if tools:
            functions: list[Function] = [x.to_chat_function() for x in tools]
            if force_tool:
                match force_tool:
                    case str():
                        # force the usage of the specified tool
                        if force_tool not in {x["name"] for x in functions}:
                            raise ValueError("Supplied force_tool value is not a valid function name")
                        function_call = {"name": force_tool}
                    case bool():
                        # use the name of the single supplied tool automatically
                        if len(tools) > 1:
                            raise ValueError("Cannot use force_tool=True shortcut when supplying more than one tool!")
                        function_call = {"name": functions[0]["name"]}
                    case _:
                        raise ValueError("Bad force_tool value")

            else:
                function_call = "auto"

            response = self.openai_client().chat.completions.create(
                model=self.model, messages=chat_history, functions=functions, function_call=function_call
            )
        else:
            response = self.openai_client().chat.completions.create(model=self.model, messages=chat_history)
        self.log.debug(response)
        return response


class FlaggedInputError(RuntimeError):
    def __init__(self, response: ModerationCreateResponse):
        super().__init__()
        self.response = response


def check_for_flagged_content(msg: str):
    response = openai.moderations.create(input=msg)

    if response.results[0].flagged:
        raise FlaggedInputError(response=response)



from streamlit.testing.v1 import AppTest


def app_script():
    import streamlit as st
    from logzero import logger

    import supersullytools.streamlit.sessions as streamlit_sst

    class TestSession(streamlit_sst.StreamlitSessionBase):
        name: str = "Guest"

    manager = streamlit_sst.DynamoDbSessionManager(
        memory=st.session_state.dynamodb_memory, model_type=TestSession, logger=logger
    )
    session = manager.init_session()
    manager.persist_session(session)


def test_basic(dynamodb_memory):
    dynamodb_memory.track_stats = False

    at = AppTest.from_function(app_script)
    at.session_state["dynamodb_memory"] = dynamodb_memory
    at.run()
    assert not at.exception
    items = dynamodb_memory.dynamodb_table.scan()["Items"]
    assert len(items) == 1



import os
from typing import TYPE_CHECKING
from uuid import uuid4

import boto3
import pytest
import requests
from logzero import logger
from simplesingletable import DynamoDbMemory
from simplesingletable.utils import truncate_dynamo_table

if TYPE_CHECKING:
    from mypy_boto3_dynamodb.service_resource import Table


@pytest.fixture()
def dynamodb_memory(local_dynamodb_test_table, dynamodb_via_docker) -> DynamoDbMemory:
    reset_local_dynamodb_test_table(local_dynamodb_test_table)
    yield DynamoDbMemory(
        logger=logger,
        table_name=local_dynamodb_test_table.table_name,
        endpoint_url=dynamodb_via_docker,
        connection_params={
            "aws_access_key_id": "unused",
            "aws_secret_access_key": "unused",
            "region_name": "us-west-2",
        },
    )


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return os.path.join(str(pytestconfig.rootdir), "src", "tests", "docker-compose.yml")


def is_responsive(url):
    try:
        response = requests.get(url)
        if response.status_code == 400:
            return True
    except requests.ConnectionError:
        return False


@pytest.fixture(scope="session")
def dynamodb_via_docker(docker_ip, docker_services):
    # `port_for` takes a container port and returns the corresponding host port
    port = docker_services.port_for("dynamodb", 8000)
    url = "http://{}:{}".format(docker_ip, port)
    docker_services.wait_until_responsive(timeout=30.0, pause=0.1, check=lambda: is_responsive(url))
    return url


@pytest.fixture(scope="session")
def local_dynamodb_test_table(dynamodb_via_docker) -> "Table":
    table_created = False
    table_name = f"delta-dynamodb-test-table-{uuid4().hex}"

    aws_access_key_id = "unused"
    aws_secret_access_key = "unused"
    region_name = "us-west-2"

    client = boto3.client(
        "dynamodb",
        endpoint_url=dynamodb_via_docker,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    resource = boto3.resource(
        "dynamodb",
        endpoint_url=dynamodb_via_docker,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )

    create_table_kwargs = {
        "TableName": table_name,
        "KeySchema": [
            {"AttributeName": "pk", "KeyType": "HASH"},
            {"AttributeName": "sk", "KeyType": "RANGE"},
        ],
        "AttributeDefinitions": [
            {"AttributeName": "pk", "AttributeType": "S"},
            {"AttributeName": "sk", "AttributeType": "S"},
            {"AttributeName": "gsitype", "AttributeType": "S"},
            {"AttributeName": "gsitypesk", "AttributeType": "S"},
            {"AttributeName": "gsi1pk", "AttributeType": "S"},
            {"AttributeName": "gsi2pk", "AttributeType": "S"},
            {"AttributeName": "gsi3pk", "AttributeType": "S"},
            {"AttributeName": "gsi3sk", "AttributeType": "S"},
        ],
        "GlobalSecondaryIndexes": [
            {
                "IndexName": "gsitype",
                "KeySchema": [
                    {"AttributeName": "gsitype", "KeyType": "HASH"},
                    {"AttributeName": "gsitypesk", "KeyType": "RANGE"},
                ],
                "Projection": {"ProjectionType": "ALL"},
            },
            {
                "IndexName": "gsi1",
                "KeySchema": [
                    {"AttributeName": "gsi1pk", "KeyType": "HASH"},
                    {"AttributeName": "pk", "KeyType": "RANGE"},
                ],
                "Projection": {"ProjectionType": "ALL"},
            },
            {
                "IndexName": "gsi2",
                "KeySchema": [
                    {"AttributeName": "gsi2pk", "KeyType": "HASH"},
                    {"AttributeName": "pk", "KeyType": "RANGE"},
                ],
                "Projection": {"ProjectionType": "ALL"},
            },
            {
                "IndexName": "gsi3",
                "KeySchema": [
                    {"AttributeName": "gsi3pk", "KeyType": "HASH"},
                    {"AttributeName": "gsi3sk", "KeyType": "RANGE"},
                ],
                "Projection": {"ProjectionType": "ALL"},
            },
        ],
        "BillingMode": "PAY_PER_REQUEST",
    }
    try:
        resp = client.create_table(**create_table_kwargs)
        dynamodb_table_arn = resp["TableDescription"]["TableArn"]
        waiter = client.get_waiter("table_exists")
        waiter.wait(TableName=table_name)
        logger.info(f"Dynamo Table Created {table_name=} {dynamodb_table_arn=}")
        table_created = True
        yield resource.Table(table_name)
    finally:
        if table_created:
            logger.info(f"Deleting generated dynamo table {table_name=}")
            client.delete_table(TableName=table_name)


def reset_local_dynamodb_test_table(table: "Table"):
    logger.debug(f"Resetting table {table.table_name}")
    truncate_dynamo_table(table)
